module nscronig.onig;

private import oniguruma;
private import std.string;
private import nscr.result;
private import nscronig.meddle;

/// バージョンを取得する。
/**
    @return 数値:バージョンの3桁表現、文字列:鬼車のバージョン表記
**/
Result getVersion(){
    return Result(ONIGURUMA_VERSION_MAJOR*100 + ONIGURUMA_VERSION_MINOR*10 + ONIGURUMA_VERSION_TEENY, toString(onig_version()));
}


private:
// ポインタはconstでは記述できないため、staticで初期化する。
/// ShiftJIS
static OnigEncodingType* ONIG_ENCODING_SJIS;
/// ASIS(plain-text)
static OnigSyntaxType* ONIG_SYNTAX_ASIS;

static this(){
    ONIG_ENCODING_SJIS         = (&OnigEncodingSJIS);
    ONIG_SYNTAX_ASIS           = (&OnigSyntaxASIS);
    options                    = ONIG_OPTION_DEFAULT;
}

// 保持する内容
/// [鬼車]正規表現構造体。適用中は非null。
static regex_t* regexp;
/// 正規表現文字列
static ubyte* pattern;
/// 正規表現文字列長
static int pattern_len;
/// 対象文字列
static ubyte* str;
/// 対象文字列長
static int str_len;
/// [鬼車]エラー情報
static OnigErrorInfo einfo;
/// [鬼車]マッチ領域
static OnigRegion* region;
/// 現在の領域位置
static int region_pos;
/// [鬼車]正規表現オブジェクトオプション
static uint options;

public:

/// 正規表現パターンを設定する。
/**
    @param  patstr  正規表現文字列
    @retval true    必ずtrue。
**/
bool setPattern(sjistext patstr){
    pattern_len = patstr.length;
    pattern = cast(ubyte*)(patstr);
    return true;
}


/// オプションを設定する。
/**
    @param  words   オプション項目語句
    @retval false   非対応語句あり
    @retval true    全て対応語句
    @note   API.jaより抜粋。
      ONIG_OPTION_NONE               オプションなし
      ONIG_OPTION_SINGLELINE         '^' -> '\A', '$' -> '\Z'
      ONIG_OPTION_MULTILINE          '.'が改行にマッチする
      ONIG_OPTION_IGNORECASE         曖昧マッチ オン
      ONIG_OPTION_EXTEND             パターン拡張形式
      ONIG_OPTION_FIND_LONGEST       最長マッチ
      ONIG_OPTION_FIND_NOT_EMPTY     空マッチを無視
      ONIG_OPTION_NEGATE_SINGLELINE
            ONIG_SYNTAX_POSIX_BASIC, ONIG_SYNTAX_POSIX_EXTENDED,
            ONIG_SYNTAX_PERL, ONIG_SYNTAX_PERL_NG, ONIG_SYNTAX_JAVAで
            デフォルトで有効なONIG_OPTION_SINGLELINEをクリアする。

      ONIG_OPTION_DONT_CAPTURE_GROUP 名前付き捕獲式集合のみ捕獲
      ONIG_OPTION_CAPTURE_GROUP      名前無し捕獲式集合も捕獲
**/
bool setOption(char[][] words){
    uint opts = ONIG_OPTION_NONE;
    bool success = true;
    foreach(word; words){
        switch(tolower(word)){
        case "s":
        case "singleline":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_SINGLELINE);
            break;
        case "m":
        case "multiline":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_MULTILINE);
            break;
        case "i":
        case "ignorecase":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_IGNORECASE);
            break;
        case "x":
        case "extend":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_EXTEND);
            break;
        case "l":
        case "findlongest":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_FIND_LONGEST);
            break;
        case "n":
        case "findnotempty":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_FIND_NOT_EMPTY);
            break;
        case "dontcapturegroup":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_DONT_CAPTURE_GROUP);
            break;
        case "capturegroup":
            opts = ONIG_OPTION_ON(opts, ONIG_OPTION_CAPTURE_GROUP);
            break;
        // "g"のみ大文字小文字を区別する。
        case "g":
            opts = ONIG_OPTION_ON(opts, ((word == "g") ? ONIG_OPTION_DONT_CAPTURE_GROUP : ONIG_OPTION_CAPTURE_GROUP));
            break;
        default:
            success = false;
            break;
        }
    }
    options = opts;
    return success;
}


/// 文章内を検索する。
/**
    @retval -1  適合する文字列がない
    @retval -2  エラー発生
    @retval >=0 数値:適合文字列の先頭位置、文字列:適合文字列
**/
RawResult find(sjistext text){
    int r;
    if(regexp is null){
        // ONIG_SYNTAX_ASISにするとまともにマッチしてくれない・・・
        r = onig_new(&regexp, pattern, pattern + pattern_len,
            options, ONIG_ENCODING_SJIS, ONIG_SYNTAX_DEFAULT, &einfo);
        if (r != ONIG_NORMAL) {
            char s[ONIG_MAX_ERROR_MESSAGE_LEN];
            onig_error_code_to_str(cast(ubyte*)s, r, &einfo);
            return RawResult(-2, "ERROR: " ~ s.dup);
        }
    
        region = onig_region_new();
    }

    // 以下はsimple.cより。

    ubyte* start;
    ubyte* range;
    ubyte* end;

    str_len = text.length;
    str = cast(ubyte*)(text);

    end   = str + str_len;
    start = str;
    range = end;
    char[] resultstr;
    r = onig_search(regexp, str, end, start, range, region, ONIG_OPTION_NONE);

    if(r >= 0){
        region_pos = 1;
        return RawResult(region.beg[0], cast(char[])str[region.beg[0]..region.end[0]]);
    }
    else if (r == ONIG_MISMATCH) {
        return RawResult(-1, "");
    }
    else{
        char s[ONIG_MAX_ERROR_MESSAGE_LEN];
        onig_error_code_to_str(cast(ubyte*)s, r, &einfo);
        return RawResult(-2, "ERROR: " ~ s.dup);
    }
}


/// 次の文章を検索する。
/**
    @retval -1  一度も検索してない or 検索が最後尾までいった
    @retval >=0 数値:適合文字列の先頭位置、文字列:適合文字列
**/
RawResult next(){
    if(region is null){
        return RawResult(-1, "");
    }
    if(region_pos >= region.num_regs){
        return RawResult(-1, "");
    }

    int pos = region_pos;
    region_pos++;
    return RawResult(region.beg[pos], cast(char[])str[region.beg[pos]..region.end[pos]]);
}


/// 検索を終了する。
/**
    @retval false   既に終了済み
    @retval true    検索を終了
    @note   正規表現文字列、検索対象文字列は保持したまま。
**/
bool end(){
    if(regexp is null) return false;
    onig_region_free(region, 1 /* 1:free self, 0:free contents only */);
    onig_free(regexp);
    onig_end();
    region = null;
    regexp = null;
    return true;
}
